@@ -1,10 +1,11 @@
-use lsp_types::{Position, Range, TextDocumentContentChangeEvent};
+use lsp_types::{Position, PositionEncodingKind, Range, TextDocumentContentChangeEvent};
 
 #[derive(Debug)]
 pub struct FullTextDocument {
     language_id: String,
     version: i32,
     content: String,
+    encoding: PositionEncodingKind,
 
     /// The value at index `i` in `line_offsets` is the index into `content`
     /// that is the start of line `i`. As such, the first element of
@@ -57,13 +58,168 @@ fn line_offset_utf16(line: &str, offset: u32) -> u32 {
     c
 }
 
+fn line_offset_utf8(line: &str, offset: u32) -> u32 {
+    offset.min(
+        line
+            .len()
+            .try_into()
+            .expect("The length of the text involved in the calculation is too long"),
+    )
+}
+
+fn line_offset_utf32(line: &str, offset: u32) -> u32 {
+    let mut c = 0;
+    for (idx, ch) in line.char_indices() {
+        if idx + ch.len_utf8() > offset as usize || idx == offset as usize {
+            break;
+        }
+        c += 1;
+    }
+    c
+}
+
+fn units_to_byte_offset_utf16(line: &str, character: u32) -> u32 {
+    let mut c = 0;
+    for (idx, ch) in line.char_indices() {
+        if c == character {
+            return idx as u32;
+        }
+        let next = c + ch.len_utf16() as u32;
+        if next > character {
+            return idx as u32;
+        }
+        c = next;
+    }
+    line.len() as u32
+}
+
+fn units_to_byte_offset_utf8(line: &str, character: u32) -> u32 {
+    character.min(
+        line
+            .len()
+            .try_into()
+            .expect("The length of the text involved in the calculation is too long"),
+    )
+}
+
+fn units_to_byte_offset_utf32(line: &str, character: u32) -> u32 {
+    let mut c = 0;
+    for (idx, _ch) in line.char_indices() {
+        if c == character {
+            return idx as u32;
+        }
+        c += 1;
+        if c > character {
+            return idx as u32;
+        }
+    }
+    line.len() as u32
+}
+
+fn bytes_to_units(line: &str, offset: u32, encoding: &PositionEncodingKind) -> u32 {
+    if encoding == &PositionEncodingKind::UTF8 {
+        line_offset_utf8(line, offset)
+    } else if encoding == &PositionEncodingKind::UTF16 {
+        line_offset_utf16(line, offset)
+    } else {
+        line_offset_utf32(line, offset)
+    }
+}
+
+fn units_to_bytes(line: &str, character: u32, encoding: &PositionEncodingKind) -> u32 {
+    if encoding == &PositionEncodingKind::UTF8 {
+        units_to_byte_offset_utf8(line, character)
+    } else if encoding == &PositionEncodingKind::UTF16 {
+        units_to_byte_offset_utf16(line, character)
+    } else {
+        units_to_byte_offset_utf32(line, character)
+    }
+}
+
 impl FullTextDocument {
     pub fn new(language_id: String, version: i32, content: String) -> Self {
+        Self::new_with_encoding(language_id, version, content, PositionEncodingKind::UTF16)
+    }
+
+    /// Create a new text document with a specific position encoding
+    ///
+    /// This method allows you to create a text document with a specific position encoding
+    /// for character positions. The encoding determines how character offsets are calculated
+    /// and is important for proper LSP communication between client and server.
+    ///
+    /// Use this method instead of [`new()`](Self::new) when you need to use a position encoding
+    /// other than UTF-16, or when you want to explicitly specify the encoding to match what
+    /// was negotiated with the LSP client during initialization.
+    ///
+    /// # Arguments
+    ///
+    /// * `language_id` - The language identifier for the document (e.g., "rust", "javascript")
+    /// * `version` - The version number of the document
+    /// * `content` - The full text content of the document
+    /// * `encoding` - The position encoding to use. Can be UTF-8, UTF-16, or UTF-32.
+    ///
+    /// # Position Encodings
+    ///
+    /// - **UTF-16**: The default encoding for backward compatibility with LSP 3.16 and earlier.
+    ///   Each UTF-16 code unit counts as one position unit. Use [`new()`](Self::new) for this.
+    /// - **UTF-8**: Each byte counts as one position unit. More efficient for ASCII-heavy text.
+    /// - **UTF-32**: Each Unicode code point counts as one position unit.
+    ///
+    /// # Examples
+    ///
+    /// Basic usage with UTF-16 (default):
+    ///
+    /// ```
+    /// use lsp_textdocument::FullTextDocument;
+    /// use lsp_types::PositionEncodingKind;
+    ///
+    /// let doc = FullTextDocument::new_with_encoding(
+    ///     "rust".to_string(),
+    ///     1,
+    ///     "fn main() {}".to_string(),
+    ///     PositionEncodingKind::UTF16
+    /// );
+    /// ```
+    ///
+    /// Using UTF-8 encoding for better performance with ASCII text:
+    ///
+    /// ```
+    /// use lsp_textdocument::FullTextDocument;
+    /// use lsp_types::PositionEncodingKind;
+    ///
+    /// let doc = FullTextDocument::new_with_encoding(
+    ///     "javascript".to_string(),
+    ///     1,
+    ///     "console.log('Hello');".to_string(),
+    ///     PositionEncodingKind::UTF8
+    /// );
+    /// ```
+    ///
+    /// Using UTF-32 encoding where each Unicode code point is one unit:
+    ///
+    /// ```
+    /// use lsp_textdocument::FullTextDocument;
+    /// use lsp_types::PositionEncodingKind;
+    ///
+    /// let doc = FullTextDocument::new_with_encoding(
+    ///     "plain".to_string(),
+    ///     1,
+    ///     "Hello 🦀 World".to_string(),
+    ///     PositionEncodingKind::UTF32
+    /// );
+    /// ```
+    pub fn new_with_encoding(
+        language_id: String,
+        version: i32,
+        content: String,
+        encoding: PositionEncodingKind,
+    ) -> Self {
         let line_offsets = computed_line_offsets(&content, true, None);
         Self {
             language_id,
             version,
             content,
+            encoding,
             line_offsets,
         }
     }
@@ -161,6 +317,12 @@ impl FullTextDocument {
         &self.language_id
     }
 
+    /// Returns the position encoding kind used by this document when
+    /// converting between LSP positions and byte offsets.
+    pub fn encoding(&self) -> PositionEncodingKind {
+        self.encoding.clone()
+    }
+
     /// Document's version
     pub fn version(&self) -> i32 {
         self.version
@@ -239,7 +401,7 @@ impl FullTextDocument {
             // only one line
             return Position {
                 line: 0,
-                character: line_offset_utf16(self.get_line(0).unwrap(), offset),
+                character: bytes_to_units(self.get_line(0).unwrap(), offset, &self.encoding),
             };
         }
 
@@ -262,7 +424,7 @@ impl FullTextDocument {
             // offset is on the first line
             return Position {
                 line: 0,
-                character: line_offset_utf16(self.get_line(0).unwrap(), offset),
+                character: bytes_to_units(self.get_line(0).unwrap(), offset, &self.encoding),
             };
         }
 
@@ -270,9 +432,10 @@ impl FullTextDocument {
 
         Position {
             line,
-            character: line_offset_utf16(
+            character: bytes_to_units(
                 self.get_line(line).unwrap(),
                 offset - self.line_offsets[line as usize],
+                &self.encoding,
             ),
         }
     }
@@ -283,15 +446,7 @@ impl FullTextDocument {
         let Position { line, character } = position;
         match self.get_line_and_offset(line) {
             Some((line, offset)) => {
-                let mut c = 0;
-                let iter = line.char_indices();
-                for (idx, char) in iter {
-                    if c == character {
-                        return offset + idx as u32;
-                    }
-                    c += char.len_utf16() as u32;
-                }
-                offset + line.len() as u32
+                offset + units_to_bytes(line, character, &self.encoding)
             }
             None => {
                 if line >= self.line_count() {
@@ -307,6 +462,7 @@ impl FullTextDocument {
 #[cfg(test)]
 mod tests {
     use super::*;
+    use lsp_types::PositionEncodingKind;
 
     fn full_text_document() -> FullTextDocument {
         FullTextDocument::new(
@@ -386,6 +542,83 @@ mod tests {
         assert_eq!(offset, 8);
     }
 
+    #[test]
+    fn test_utf8_encoding_positions() {
+        let text_document = FullTextDocument::new_with_encoding(
+            "plain".to_string(),
+            1,
+            "\u{1F496}abc".to_string(),
+            PositionEncodingKind::UTF8,
+        );
+
+        // UTF-8 character offset is in bytes; emoji is 4 bytes.
+        assert_eq!(
+            text_document.offset_at(Position {
+                line: 0,
+                character: 4,
+            }),
+            4
+        );
+        assert_eq!(
+            text_document.position_at(4),
+            Position {
+                line: 0,
+                character: 4,
+            }
+        );
+
+        // beyond line end clamps to line end
+        assert_eq!(
+            text_document.offset_at(Position {
+                line: 0,
+                character: 99,
+            }),
+            text_document.content_len()
+        );
+    }
+
+    #[test]
+    fn test_utf32_encoding_positions() {
+        let text_document = FullTextDocument::new_with_encoding(
+            "plain".to_string(),
+            1,
+            "\u{1F496}abc".to_string(),
+            PositionEncodingKind::UTF32,
+        );
+
+        // UTF-32 character offset is in Unicode scalars; emoji counts as 1.
+        assert_eq!(
+            text_document.offset_at(Position {
+                line: 0,
+                character: 1,
+            }),
+            4
+        );
+        assert_eq!(
+            text_document.position_at(4),
+            Position {
+                line: 0,
+                character: 1,
+            }
+        );
+
+        // Position after the emoji should move into ASCII bytes/scalars correctly.
+        assert_eq!(
+            text_document.offset_at(Position {
+                line: 0,
+                character: 3,
+            }),
+            6
+        );
+        assert_eq!(
+            text_document.position_at(6),
+            Position {
+                line: 0,
+                character: 3,
+            }
+        );
+    }
+
     #[test]
     fn test_position_at() {
         let text_document = full_text_document();
